#!/usr/bin/awk -f
#
# awkiawki - wikiwiki clone written in (n|g|m)awk
#
# Copyright (c) 2002 Oliver Tonnhofer <olt@bogosoft.com>
# Copyright (c) 2003 Paul H. Liu <paul@thev.net>
#
# See the file `COPYING' for copyright notice.

BEGIN {
    ###########################################################
    #            --- default options ---
    #    --- use awki.conf to override default settings ---
    #
    #    datadir: Directory for raw pagedata (must be writeable for the script).
    localconf["datadir"] = datadir ? datadir : "./data/"
    #    parser: Parsing script.
    localconf["parser"] = "./parser.awk"
    #   special_parser: Parser for special_* functions.
    localconf["special_parser"] = "./special_parser.awk"
    #   special_list
    localconf["special_list"] = "ls -1"
    localconf["special_list_by_time"] = "ls -tlL"
    #    default_page: Name of the default_page.
    localconf["default_page"] = "FrontPage"
    #    show_changes: Number of changes listed by RecentChanges
    localconf["show_changes"] = 10
    #    max_post: Bytes accept by POST requests (to avoid DOS).
    localconf["max_post"] = 100000
    #    write_protection: Regex for write protected files
    #       e.g.: "*", "PageOne|PageTwo|^.*NonEditable" 
    #        HINT: to edit these protected pages, upload a .htaccess 
    #              protected awki.cgi script with write_protection = ""
    localconf["write_protection"] = ""
    #   css: HTTP URL for external CSS file.
    localconf["css"] = ""
    #   always_convert_spaces: If true, convert runs of 8 spaces to tab automatical.
    localconf["always_convert_spaces"] = 0
    #    date_cmd: Command for current date.
    localconf["date_cmd"] = "date '+%e %b. %G %R:%S %Z'"
    localconf["date_fmt"] ="+%e %b. %G %R:%S %Z"
    #    header & footer
    localconf["header_html"] = "no"
    localconf["footer_html"] = "no"
    localconf["charset"] = "iso-8859-1"
    localconf["searchstr"] = ""
    #    the regexp for wiki names, note that \ shall be double quoted as \\
    localconf["wikiexp"] = "[A-Z]+[a-z]+[A-Z][A-Za-z]*|" \
                           "[A-Z][A-Z]+[a-z][A-Za-z]*|" \
                           "_(\\w|[\\\\.:_-])+_"

    #            --- default options ---
    ###########################################################

    scriptname = ENVIRON["SCRIPT_NAME"]

    #load external configfile
    load_config(scriptname)
    
    # PATH_INFO contains page name
    if (ENVIRON["PATH_INFO"]) {    
        query["page"] = ENVIRON["PATH_INFO"]
    }
    
    # process user upload via HTTP-POST
    if (ENVIRON["REQUEST_METHOD"] == "POST") {
        # normal form post
        if (ENVIRON["CONTENT_TYPE"] == "application/x-www-form-urlencoded") {
            # restrict the length to "max_post"
            if (ENVIRON["CONTENT_LENGTH"] < localconf["max_post"])
                bytes = ENVIRON["CONTENT_LENGTH"]
            else
                bytes = localconf["max_post"]
            cmd = "dd ibs=1 count=" bytes " 2>/dev/null"
            # treat the form data as normal query string
            cmd | getline query_str
            close (cmd)
            if (ENVIRON["QUERY_STRING"])
                query_str = query_str "&" ENVIRON["QUERY_STRING"]
        }
        # file upload via MIME multipart
        else if (match(ENVIRON["CONTENT_TYPE"],
                 /multipart\/form-data; boundary=/) > 0) {
            boundary = substr(ENVIRON["CONTENT_TYPE"], RLENGTH + 1)
            # dump the content into a temporary file, Unix system only
            cmd = "mktemp /tmp/temp.XXXXXX"
            cmd | getline tempfile
            close (cmd)
            # use special_parser to copy from stdin to tempfile
            cmd = "dd ibs=1 count=" ENVIRON["CONTENT_LENGTH"] \
                   " | " localconf["special_parser"] \
                   " -v special_upload=true -v boundary='" boundary \
                   "' -v tempfile='" tempfile "' 2> /dev/null"
            cmd | getline result
            close (cmd)
            
            # the output from special_parser is the upload filename and size
            # separated by \t
            gsub(/ +/, " ", result)
            match(result, /\t/)
            query["uploadfilename"] = substr(result, 1, RSTART - 1)
            query["tempfile"] = tempfile
            query["tempfile_size"] = substr(result, RSTART + 1, 
                                            length(result) - RSTART + 1)
            if (ENVIRON["QUERY_STRING"])    
                query_str = ENVIRON["QUERY_STRING"]
        }
    }
    # otherwise for HTTP-GET, the query_str is all in the environment
    else {
        if (ENVIRON["QUERY_STRING"])    
            query_str = ENVIRON["QUERY_STRING"]
    }
    
    # split the query_str into a table of key value pairs
    n = split(query_str, querys, "&")
    for (i=1; i<=n; i++) {
        split(querys[i], data, "=")
        query[data[1]] = data[2]
    }

    # (IMPORTANT for security!)
    # TODO: we may need to check more stuff here
    query["page"] = clear_pagename(query["page"])
    query["revision"] = clear_revision(query["revision"])
    query["string"] = clear_str(decode(query["string"]))
    
    if (query["page"] == "")
        query["page"] = localconf["default_page"]

    #check if page is editable, special_pages are not
    special_pages = "FullSearch|PageList|RecentChanges"
    
    if (query["page"] ~ "("special_pages")") {
        special_page = 1
    } 
    else {
        # the real file of the queried page in local filesystem
        query["filename"] = localconf["datadir"] query["page"]

        if (!localconf["write_protection"] || 
             query["page"] !~ "("localconf["write_protection"]")")
                page_editable = 1
    }

    # we shall redirect the page immediately after an edit
    # to view its updated version
    if (query["save"]) redirect = scriptname "/" query["page"]

    if (query["filename"])
    {
        # get the file's last modified time without using rcs
        cmd = localconf["special_list"] " " query["filename"]
        cmd | getline filetime
        close(cmd)
        if (match(filetime, /\t/) > 1) {
            timestr = substr(filetime, RSTART + 1)
            date = strftime(localconf["date_fmt"], timestr)
        }
    }

    # if the file exist, and query_str has no time "stamp",
    # and the command isn't any of of the special commands,
    # we shall redirect to the latest file link.
    #
    # this is a trick to make sure newest page revision can
    # be presented without forcing immediate page expiration
    # so as to save some load on server side.
    #
    if (timestr && !query["stamp"] && !query["revision"] &&
        !query["edit"] && !query["save"] && !query["newname"] && 
        !query["upload"] && !query["remove_upload"] && !query["history"] &&
        !query["attach"])
    {
        redirect = scriptname "/" query["page"] "?stamp=" timestr
    }

    header(query["page"], redirect, date)

    # process each command accordingly
    if (query["edit"] && page_editable)    
        edit(query["page"], query["filename"], query["revision"])
    else if (query["rename"] && page_editable)
        rename(query["page"], query["newname"])
    else if (query["save"] && query["text"] && page_editable)
        save(query["page"], query["text"], query["string"], query["filename"])
    else if (query["upload"] && page_editable)
        upload(query["page"], query["uploadfilename"], query["tempfile"], 
               query["tempfile_size"], query["filename"])
    else if (query["remove_upload"] && page_editable)
        remove_upload(query["page"], query["name"], query["filename"])
    else if (query["page"] ~ "PageList")
        pagelist(localconf["datadir"])
    else if (query["page"] ~ "RecentChanges")
        changes(localconf["datadir"])
    else if (query["page"] ~ "FullSearch")
        search(query["string"], query["case"], localconf["datadir"])
    else if (query["page"] && query["history"])
        history(query["page"], query["filename"])
    else if (query["page"] && query["attach"] && page_editable)
        attach(query["page"], query["filename"])
    else if (query["stamp"] || query["revision"])
        parse(query["page"], query["filename"], query["revision"])

    footer(query["page"], redirect)
    
}

# get the printed name of a page
function page_name(page, name) {
    if (match(page, /_(\w|[\\.:_-])+_/))
    {
        name = substr(page, RSTART + 1, RLENGTH - 2)
        gsub(/_/, " ", name)
    }
    else 
    {
        name = page
    }
    return name
}

# print header
function header(page, redirect, date) {
    visname = page_name(page)
    print "Content-type: text/html\n"

    print localconf["scriptname"]

    if (redirect)
    {
        print "<html><header><meta http-equiv=\"Refresh\" content=\"1;URL=" \
              redirect "\" /></header>"
    }
    else
    {
        if (!special_page) {
          edt_href = "href"
          att_href = "href"
          rev_href = "href"
        }
        else {
          edt_href = "name"
          att_href = "name"
          rev_href = "name"
        }

        while ((getline < localconf["header_html"]) >0) {
            gsub(/[$]TITLE[$]/, visname);
            gsub(/[$]CHARSET[$]/, localconf["charset"])
            gsub(/[$]SEARCH[$]/, localconf["searchstr"])
            gsub(/[$]PAGE[$]/, page)
            gsub(/[$]SCRIPTNAME[$]/, scriptname)
            gsub(/[$]DEFAULTPAGE[$]/, localconf["default_page"])
            gsub(/[$]EDT_HREF[$]/, edt_href)
            gsub(/[$]REV_HREF[$]/, rev_href)
            gsub(/[$]ATT_HREF[$]/, att_href)
            gsub(/[$]DATE[$]/, date)
            print
        }
        close(localconf["header_html"])
    }
}

# print footer
function footer(page, redirect) {
    if (redirect) {
        print "</html>"
    }
    else {
        while ((getline < localconf["footer_html"]) >0)
            print
        close(localconf["footer_html"])
    }
}

# send page to parser script
function parse(name, filename, revision) {
    # if the file really exists
    if (system("[ -f " filename" ]") == 0 ) {
        if (revision) {
            print "<em>Displaying old revision (" revision \
                  ") of <a href=\"" scriptname "/" name "\">" name \
                  "</a>.</em><p />"
            system("co -q -p'" revision "' " filename " | " \
                   localconf["parser"] " -v datadir='" localconf["datadir"] \
                   "' -v revision='" revision \
                   "' -v filename='" filename \
                   "' -v special_parser='" localconf["special_parser"] \
                   "' -v special_list='" localconf["special_list"] \
                   "' -v mypage='" name "'")
        } else {
            system(localconf["parser"] \
                   " -v datadir='"localconf["datadir"] \
                   "' -v mypage='" name \
                   "' -v special_parser='" localconf["special_parser"] \
                   "' -v special_list='" localconf["special_list"] \
                   "' " filename)
        }
    }
}

# rename a page
function rename(page, newpage) {
    cmd = "ls " datadir
    while ((cmd | getline ls_out) >0) {
        if (match(ls_out, localconf["wikiexp"]) &&
            substr(ls_out, RSTART + RLENGTH) !~ /,v/) {
            p = substr(ls_out, RSTART, RLENGTH)
            pages[p] = 1
        }
    }
    close(cmd)
    if (pages[newpage])
    {
        print "You cannot rename <font color=\"blue\">" page \
              "</font> to <font color=\"blue\">" newpage \
              "</font> because it already exists!"
    }
    else if (newpage !~ ("^" localconf["wikiexp"] "$"))
    {
        print "You cannot rename <font color=\"blue\">" page \
              "</font> to <font color=\"blue\">" newpage \
              "</font> because it is not a valid Wiki name!"
    }
    else
    {
        filename = localconf["datadir"] page
        newfilename = localconf["datadir"] newpage
        system("mv '" filename "' '" newfilename "'")

        # rename the attachment directory as well
        system("mv '" localconf["datadir"] "UPLOAD/" page \
               "' '" localconf["datadir"] "UPLOAD/" newpage "'")

        localconf["date_cmd"] | getline date
        close(localconf["date_cmd"])
        system("ci -q -t-"page" -l -m'" ENVIRON["REMOTE_ADDR"] ";;" \
               date ";;renamed from " page "' " newfilename)

        # replace old name occurrences in all pages
        system("grep -l '" page "' " localconf["datadir"] "* | " \
               localconf["special_parser"] \
               " -v special_rename=yes -v page='" page \
               "' -v newpage='" newpage \
               "' -v date='" date "' -v addr='" ENVIRON["REMOTE_ADDR"] "'")

        print "<font color=\"blue\">" page \
              "</font> is now renamed <a href=\""scriptname"/" newpage "\">" \
              newpage "</a>"
    }
}

# print edit form
function edit(page, filename, revision, cmd) {
    if (revision)
        print "<p /><em>If you save previous versions, " \
              "you'll overwrite the current page.</em>"
    print "<form action=\"" scriptname "?save=true&amp;page=" page \
          "\" method=\"POST\">"

    # this is a trick not to leave a newline at the beginning of the
    # form text area
    old_ORS = ORS
    ORS = ""
    print "<textarea name=\"text\" rows=25 cols=80>"
    ORS = old_ORS

    # insert current page into textarea
    if (revision) {
        cmd = "co -q -p'"revision"' " filename
        system(cmd)
    } else {
        while ((getline < filename) >0)
            print
        close(filename)
    }
    print "</textarea><br />"
    print "<input type=\"submit\" value=\"save\">"
    print "Comment: <input type=\"text\" name=\"string\" maxlength=80 size=50>"
    if (! localconf["always_convert_spaces"])
        print "<br>Convert runs of 8 spaces to Tab <input type=\"checkbox\" \
              name=\"convertspaces\">"
    print "</form>"
    print "<small><strong>Emphases:</strong> ''<em>italic</em>''; \
          '''<strong>bold</strong>'''; \
          '''''<strong><em>bold italic</em></strong>'''''; \
          ''<em>mixed '''<strong>bold</strong>'''\
          and italic</em>''<br>\
          <strong>Horizontal Rule:</strong> ----<br><strong>\
          Paragraph:</strong> blank line<br>\
          <strong>Headings:</strong> -Title 1 ; --Title 2 ; ---Title 3<br>\
          <strong>Preformatted Text:</strong> *space*Text<br>\
          <strong>Lists:</strong> tab(s) and one of * bullets; \
                  1 numbered items<br>\
          <strong>Links:</strong> JoinCapitalizedWords, or <br>\
          <strong>Links:</strong> url \
          (http, https, ftp, gopher, mailto, news), or<br> \
          <strong>Links:</strong> [[ url ][ name ]] \
          <form action=\"" scriptname "?rename=true&amp;page=" page \
          "\" method=\"POST\">\
          If you want to rename the page, please input a new Wiki name: \
          <input type=\"text\" name=\"newname\" maxlength=30 />\
          <input type=\"submit\" value=\"rename\"></form><br>"
}

# save page after a form submission from edit
function save(page, text, comment, filename, dtext, date) {
    dtext = decode(text);
    if ( localconf["always_convert_spaces"] || query["convertspaces"] == "on")
        gsub(/        /, "\t", dtext);
    print dtext > filename
    localconf["date_cmd"] | getline date
    close(localconf["date_cmd"])
    system("ci -q -t-"page" -l -m'"ENVIRON["REMOTE_ADDR"] ";;" date \
           ";;"comment"' " filename)

    print "The editing of <a href=\""scriptname"/"page"\">"page"</a> is saved."
}

# list all page attachments
function attach(page, filename) {
    print "<h3>All attached files to this page are listed below</h3>" \
          "<table border=1 cellpadding=4><tr bgcolor=lightgreen>" \
          "<th>file name</th><th>size (bytes)</th><th>remove</th></tr>"

    system("ls -l " localconf["datadir"] "UPLOAD/" page \
           " | " localconf["special_parser"] \
           " -v special_uploadlist=true -v datadir='" localconf["datadir"] \
           "' -v page='" page "'")

    print "</table>" \
          "<form action=\"" scriptname "?upload=true&page=" page \
          "\" method=\"POST\" enctype=\"multipart/form-data\">" \
          "Please select a file from your local disk to upload<br>" \
          "<input type=file name=uploadfile size=30 maxlength=128><br>" \
          "<input type=\"submit\" value=\"Upload\">" \
          "</form>"
}

# upload file

#
function upload(page, upload_filename, tempfile, size, filename, existed) {

    # upload_filename must be a valid file name for security reasons!
    # NOTE: clear_str restricts the string length to 80 characters
    upload_filename = clear_str(upload_filename)

    if ((filename == "") || (size == 0)) {
        print "Please specify a valid file to upload.<br>"
    } else {
        cmd = "test -f '" localconf["datadir"] "UPLOAD/" page "/" 
                upload_filename "' && echo 1"
        cmd | getline existed
        close(cmd)
      
        if (existed == 1) {
            print "File " upload_filename " already exists, please " \
                  " upload another file or remove the previous one first.<p>"
        } else {
            system("mkdir -m 755 -p '" localconf["datadir"] "UPLOAD/" page "'")
            system("chmod go+r " tempfile)
            system("mv '" tempfile "' '" localconf["datadir"] "UPLOAD/" page \
                   "/" upload_filename "'")
            print "File " upload_filename " is successfully uploaded.<p>"
        }
    }
    # after the upload, just display all attachments
    attach(page, filename)
}

# remove uploaded attachments

function remove_upload(page, name, filename) {

   # the name must be valid filename for security reasons!!
    name = clear_str(decode(name))
    code = system("rm '" localconf["datadir"] "UPLOAD/" page "/" name "'")
    if (code == 0) 
        print "File " name " is removed.<p>"
    else
        print "Error occurred when removing File " name ".<p>"

    attach(page, filename)
}

# list all pages (sorted by name)
function pagelist(datadir) {
    system(localconf["special_list"] " " datadir " |sort| " \
           localconf["special_parser"] " -v special_index=yes")
}

# list pages with last modified time (sorted by date)
function changes(datadir, date) {
    localconf["date_cmd"] | getline date
    close(localconf["date_cmd"])
    print "<p>current date:", date "<p>"
    system(localconf["special_list_by_time"] " " datadir " | " \
           localconf["special_parser"] \
           " -v special_changes=" localconf["show_changes"] \
           " -v date_fmt='" localconf["date_fmt"] "'")
}

# keyword search
function search(name, case, datadir) {
    if (case ~ /sensitive/) { cmd = "grep -l" }
    else { cmd = "grep -il" }
    system(cmd " '"name"' "datadir"* | " \
           localconf["special_parser"] " -v special_search=yes")
}

# revision history of a page
function history(name, filename) {
    print "<p>last changes on <a href=\""scriptname"/" name \
          "\">"name"</a><p>"
    system("rlog " filename " | " localconf["special_parser"] \
           " -v special_history="name)
}

# remove '"`\/ characters from string
# *** !Important for Security! ***
function clear_str(str) {
    gsub(/['`"\/\\]/, "", str)
    if (length(str) > 80)
        return substr(str, 1, 80)
    else
        return str
}

# clean up pagename string
# *** !Important for Security! ***
function clear_pagename(str) {
    if (match(str, localconf["wikiexp"]) > 0)
        return substr(str, RSTART, RLENGTH)
    else
        return ""
}

# clean up revision numbers
# *** !Important for Security! ***
function clear_revision(str) {
    if (match(str, /[1-9]\.[0-9]+/))
        return substr(str, RSTART, RLENGTH)
    else
        return ""
}

# decode urlencoded string
function decode(text, hex, i, hextab, decoded, len, c, c1, c2, code) {
    
    split("0 1 2 3 4 5 6 7 8 9 a b c d e f", hex, " ")
    for (i=0; i<16; i++) hextab[hex[i+1]] = i

    # urldecode function from Heiner Steven
    # http://www.shelldorado.com/scripts/cmds/urldecode

    # decode %xx to ASCII char 
    decoded = ""
    i = 1
    len = length(text)
    
    while ( i <= len ) {
        c = substr (text, i, 1)
        if ( c == "%" ) {
            if ( i+2 <= len ) {
                c1 = tolower(substr(text, i+1, 1))
                c2 = tolower(substr(text, i+2, 1))
                if ( hextab [c1] != "" || hextab [c2] != "" ) {
                    if ( (c1 >= 2 && !(c1 == 7 && c2 == "F")) || 
                         (c1 == 0 && c2 ~ "[9acd]") ) {
                        code = 0 + hextab [c1] * 16 + hextab [c2] + 0
                        c = sprintf ("%c", code)
                    } else {
                        c = " "
                    }
                    i = i + 2
               }
            }
        } else if ( c == "+" ) {    # special handling: "+" means " "
            c = " "
        }
        decoded = decoded c
        ++i
    }
    
    # change linebreaks to \n
    gsub(/\r\n/, "\n", decoded)
    
    # remove last linebreak
    sub(/[\n\r]*$/,"",decoded)
    
    return decoded
}

# load configfile
function load_config(script, configfile, key, value) {
    configfile = script
    #remove trailing / ('/awki/awki.cgi/' -> '/awki/awki.cgi')
    sub(/\/$/, "", configfile)
    #remove path ('/awki/awki.cgi' -> 'awki.cgi')
    sub(/^.*\//, "", configfile)
    #remove suffix ('awki.cgi' -> 'awki')
    sub(/\.[^.]*$/,"", configfile)
    # append .conf suffix
    configfile = configfile ".conf"
    
    #read configfile
    while((getline < configfile) >0) {
        if ($0 ~ /^#/) continue #ignore comments
        
        if (match($0,/[ \t]*=[ \t]*/)) {
            key = substr($0, 1, RSTART-1)
            sub(/^[ \t]*/, "", key)
            value = substr($0, RSTART+RLENGTH)
            sub(/[ \t]*$/, "", value)
            if (sub(/^"/, "", value))
                sub(/"$/, "", value) 
            #set localconf variables
            localconf[key] = value
        
        }
    }
    close(configfile)
}    
